# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

import os as os
from pathlib import Path as Path
import cupy as cp
from ....utils._kernels import get_kernel

path = Path(__file__).parent / "sipm_signals.cu"
with open(path) as _source_file:
    source_code = _source_file.read()

include_path = ''.join([os.environ['CPATH']])
module = cp.RawModule(code=source_code, backend='nvcc', options=(''.join(['-I',include_path]),'--use_fast_math', '-std=c++11', '--extra-device-vectorization', '-O3'))

sipm_signals = get_kernel(module, 'sipm_signals')


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

from ._surface_misc import (
    ApertureShape,
    SurfaceProperties,
    SurfaceType,
    SurfaceShape,
    SurfaceTextures,
    FresnelSurfacePropertiesGenerator,
)
from ._materials import Materials, RefractiveIndexTexture
from ._surface import (
    AsphericalSurface,
    CylindricalSurface,
    FlatSurface,
    SphericalSurface,
    SipmTileSurface,
)
from ._optical_system import OpticalSystem
from ._atmosphere import AtmosphericTransmission
from ._multilayer_filter import (
    OpticalStackSimulator,
    SipmStackSimulator
)
from . import sources
from . import transforms


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import vtk

from ._vtk_utils import (
    create_aspheric_surface_actor,
    create_cylindric_surface_actor,
    quit
)

from ..optics._surface_misc import SurfaceType, SurfaceShape

from ..optics._cpu_transforms import local_to_telescope_rotation, local_to_pointing_rotation

from .._iact import IACT

from ..optics._optical_system import OpticalSystem


class VTKOpticalSystem():
    """Class to viasualize the geometry of an optical system.
    Each surface actor can be accessed from :py:attr:`actors` attribute after a :py:meth:`update` call.

    Parameters
    ----------
    optical_system : OpticalSystem or IACT
        Optical sistem for which visualize surfaces.
    resolution : float, optional
            Objects mesh resolution (in mm). By default 10 mm.

    Notes
    -----
    If you perform some operation on an actor, make sure to call :py:meth:`start_render` with ``update=False``, otherwise the actors will be replaced.
    
    """
    def __init__(self, optical_system, resolution=None):
        self.actors = {}
        """Dictionary of surface name: surface actor."""

        if issubclass(type(optical_system), OpticalSystem):
            self.os = optical_system
            self._tel_to_local_rot =  np.eye(3)
            self._translation_vector = np.zeros(3)
        elif issubclass(type(optical_system), IACT):
            self.os = optical_system.optical_system
            self._tel_to_local_rot =  local_to_telescope_rotation(*optical_system.pointing).T
            self._translation_vector = optical_system.position
        else:
            raise(ValueError("optical system must be an instance of OpticalSystem or IACT."))

        """Optical system to be visualized."""
        
        self._default_surface_color = (0.7, 0.7, 0.75)

        self.surface_type_colors = {
            SurfaceType.REFRACTIVE: (0.6, 0.4, 1),
            SurfaceType.OPAQUE: (1, 0.3, 0.3)
        }

        self.surface_type_opacity = {
            SurfaceType.REFRACTIVE: 0.3,
        }
        
        self.surface_type_specular = {
            SurfaceType.REFRACTIVE: 0.2,
            SurfaceType.REFLECTIVE: 0.8,
            SurfaceType.REFLECTIVE_SENSITIVE: 0.6
        }
        
        """Dictionary to custumize surface colors based on surface type."""

        # Wireframe
        self.wireframe = False
        """Whether to use wireframe representation by default."""

        # Window size
        self.window_size = (1024,1024)
        """Window size in pixel."""

        self._resolution = resolution # mm
        """Objects mesh resolution (in mm)."""

        self._ray_opacity = None
        """Ray opacity."""

        self._update()

        self._apply_global_transform()
    
    def _update(self):
        """Generate all surface actors.
        """
        self.actors = {}
        for i,s in enumerate(self.os):
            if s._shape == SurfaceShape.CYLINDRICAL:
                actor = create_cylindric_surface_actor(s, self._resolution)
            else:
                actor = create_aspheric_surface_actor(s, self._resolution)
                

            transform = vtk.vtkTransform()
            transform.Translate(*s.position)
            R = s.get_rotation_matrix()
            vtk_rotation_matrix = vtk.vtkMatrix4x4()
            for i in range(3):
                for j in range(3):
                    vtk_rotation_matrix.SetElement(i, j, R[j, i])
            transform.PreMultiply()
            transform.Concatenate(vtk_rotation_matrix)
            transform.Update()
            actor.SetUserTransform(transform)
            
            if self.wireframe:
                actor.GetProperty().SetRepresentationToWireframe()
            
            if s.type in self.surface_type_colors:
                color = self.surface_type_colors[s.type]
            else:
                color = self._default_surface_color
            
            if s.type in self.surface_type_opacity:
                opacity = self.surface_type_opacity[s.type]
                actor.GetProperty().SetOpacity(opacity)
            else:
                actor.GetProperty().SetOpacity(1.0)

            actor.GetProperty().SetColor(*color)
            actor.GetProperty().SetAmbient(0.3)
            actor.GetProperty().SetDiffuse(0.5)
            actor.GetProperty().SetSpecular(0.2)
            actor.GetProperty().SetSpecularPower(10)

            self.actors[s.name] = actor

    @staticmethod
    def _get_sigmoid_opacity(n_rays, n_50=1000, k=1., min_opacity=0.01):
        """
        Calculates ray opacity using a descending sigmoid curve.
        
        Parameters
        ----------
        n_rays : int
            Number of rays being plotted.
        n_50 : int
            The 'knee' of the curve: number of rays where opacity is ~0.5.
        k : float
            Steepness of the curve (the greater the steeper the curve). 
        min_opacity : float
            Minimum visibility floor.
        """
        if n_rays <= 0: return 0.0
        
        # Calculate sigmoid
        opacity = min_opacity + (1.0 - min_opacity) / (1 + (n_rays / n_50)**k)
        
        return opacity

    def start_render(self, camera_position=None, focal_point=None, view_up=(0, 0, 1), orthographic=False):
        """Render the optical system geometry on a VTK window.

        Parameters
        ----------
        camera_position : tuple or list, optional
            (x, y, z) coordinates for the camera position.
        focal_point : tuple, list, or str, optional
            If tuple/list: (x, y, z) coordinates to look at.
            If str: The name of the surface in self.os to look at.
            If None: Defaults to center of system.
        view_up : tuple or list, optional
            (x, y, z) vector defining the "up" direction. Default is (0, 0, 1).
        orthographic : bool, optional
            If True, start with a parallel projection. 
            If False, start with a perspective projection. Default is False.
        """

        # Set opacity based on number of ray actors
        # N.B.: this overestimates the number of photons.
        # N.B.: _set_ray_opacity will not update _ray_opacity
        #       at each start render the opacity will be set automatically
        #       unless an opacity is specified with set_ray_opacity
        if self._ray_opacity is None:
            n_rays = 0
            for actor in self.actors:
                if actor.startswith('rays'):
                    n_rays += self.actors[actor].GetMapper().GetDataSetInput().GetLines().GetNumberOfCells()
            opacity = self._get_sigmoid_opacity(n_rays)
            self._set_ray_opacity(opacity)

        # Rendering
        renderer = vtk.vtkRenderer()
        renderer.SetBackground(0.1, 0.15, 0.25)

        inf = float('inf')
        sys_bounds = [inf, -inf, inf, -inf, inf, -inf]
        has_system_actors = False

        for name, actor in self.actors.items():
            renderer.AddActor(actor)
            
            # Include this actor in the bounds calculation
            if not name.startswith("rays"):
                has_system_actors = True
                b = actor.GetBounds() # (xmin, xmax, ymin, ymax, zmin, zmax)
                sys_bounds[0] = min(sys_bounds[0], b[0])
                sys_bounds[1] = max(sys_bounds[1], b[1])
                sys_bounds[2] = min(sys_bounds[2], b[2])
                sys_bounds[3] = max(sys_bounds[3], b[3])
                sys_bounds[4] = min(sys_bounds[4], b[4])
                sys_bounds[5] = max(sys_bounds[5], b[5])

        # Display instructions
        desc = [
            "Press: 'q' to exit",
            "Press: 'r' to reset the camera zoom",
            "Press: 'o' to toggle orthographic/perspective view",
            "Press: 'l' to select/deselect actor under cursor",
            "Press: 'c' to move the focal point on selected actors",
            "Press: 'h' to hide the selected actors", 
            "Press: 'i' to isolate selected actors (and hide others)",
            "Press: 'Esc' to clear selection and actions",
            "Press: 'x', 'y' or 'z' to align the up-vector to the desired axis",
            "Press: 'w', 's' or 'p' to switch to wireframe, surface or points representation",
        ]
        textActor = vtk.vtkTextActor()
        textActor.SetInput('\n'.join(desc))
        position_coordinate = textActor.GetPositionCoordinate()
        position_coordinate.SetCoordinateSystemToNormalizedViewport()
        position_coordinate.SetValue(0.01, 0.99)
        textActor.GetTextProperty().SetJustificationToLeft()
        textActor.GetTextProperty().SetVerticalJustificationToTop()
        textActor.GetTextProperty().SetFontSize(18)
        textActor.GetTextProperty().SetColor(vtk.vtkNamedColors().GetColor3d("Gold"))
        textActor.GetTextProperty().SetFontFamily(vtk.VTK_COURIER)
        textActor.SetVisibility(False)
        renderer.AddActor2D(textActor)

        hintActor = vtk.vtkTextActor()
        hintActor.SetInput("Press '?' for instructions")
        hprop = hintActor.GetTextProperty()
        hprop.SetFontSize(16)
        hprop.SetColor(vtk.vtkNamedColors().GetColor3d("Gold"))
        hprop.SetFontFamily(vtk.VTK_COURIER)
        hprop.SetJustificationToLeft()
        hprop.SetVerticalJustificationToTop()
        hcoord = hintActor.GetPositionCoordinate()
        hcoord.SetCoordinateSystemToNormalizedViewport()
        hcoord.SetValue(0.01, 0.99)
        hintActor.SetVisibility(True)
        renderer.AddActor2D(hintActor)

        render_window = vtk.vtkRenderWindow()
        render_window.SetWindowName(self.os.name)
        render_window.AddRenderer(renderer)
        render_window.SetSize(*self.window_size)
        render_window_interactor = vtk.vtkRenderWindowInteractor()
        render_window_interactor.SetRenderWindow(render_window)

        # Camera
        cam_style = vtk.vtkInteractorStyleTrackballCamera()
        render_window_interactor.SetInteractorStyle(cam_style)

        # Axes
        axes = vtk.vtkAxesActor()
        widget = vtk.vtkOrientationMarkerWidget()
        rgba = [0] * 4
        vtk.vtkNamedColors().GetColor('Carrot', rgba)
        widget.SetOutlineColor(rgba[0], rgba[1], rgba[2])
        widget.SetOrientationMarker(axes)
        widget.SetInteractor(render_window_interactor)
        widget.SetEnabled(1)
        widget.InteractiveOn()

        selected_actors = {}
        is_isolated = False
        picker = vtk.vtkCellPicker()
        picker.SetTolerance(0.005)

        def perform_view_reset(up_vector):
            active_camera = renderer.GetActiveCamera()
            if not active_camera:
                return

            # Normalize up_vector
            norm_up = np.sqrt(sum(x*x for x in up_vector))
            if norm_up == 0: return
            target_up = [x/norm_up for x in up_vector]

            # Get current view plane normal
            vpn = active_camera.GetViewPlaneNormal()
            
            # Compute dot product
            dot = sum(vpn[i] * target_up[i] for i in range(3))
            
            # If dot product is near 1 or -1, vectors are parallel.
            if abs(dot) > 0.95:
                fp = active_camera.GetFocalPoint()
                dist = active_camera.GetDistance()
                
                # Move camera to a perpendicular axis based on target UP
                if abs(target_up[2]) > 0.9: 
                    active_camera.SetPosition(fp[0], fp[1] - dist, fp[2])
                elif abs(target_up[1]) > 0.9:
                    active_camera.SetPosition(fp[0] - dist, fp[1], fp[2])
                else:
                    active_camera.SetPosition(fp[0], fp[1], fp[2] + dist)

                if has_system_actors:
                    renderer.ResetCamera(sys_bounds)
                else:
                    renderer.ResetCamera()

            active_camera.SetViewUp(*up_vector)
            active_camera.OrthogonalizeViewUp() 
            render_window.Render()

        def key_press_callback(caller, event):
            interactor = caller
            key_sym = interactor.GetKeySym()
            
            if key_sym is None:
                return
            
            # Reset view ('r')
            if key_sym.lower() == 'r':
                if has_system_actors:
                    renderer.ResetCamera(sys_bounds)
                else:
                    renderer.ResetCamera()
                render_window.Render()
                interactor.SetKeySym("") 
                interactor.SetKeyCode("\0")
            
            # Instructions ('?')
            elif key_sym.lower() == 'question':
                current_vis = textActor.GetVisibility()
                # Toggle: If currently visible, make invisible.
                textActor.SetVisibility(not current_vis)
                # Hint is always the opposite of Instructions
                hintActor.SetVisibility(current_vis)
                render_window.Render()

            # Orthographic/perspective ('o')
            elif key_sym.lower() == 'o':
                cam = renderer.GetActiveCamera()
                if cam.GetParallelProjection():
                    # Switching Parallel -> Perspective
                    scale = cam.GetParallelScale()
                    fov = cam.GetViewAngle()
                    if fov > 0:
                        target_dist = scale / np.tan(np.deg2rad(fov) / 2.0)
                        fp = np.array(cam.GetFocalPoint())
                        pos = np.array(cam.GetPosition())
                        direction = fp - pos
                        dist_mag = np.linalg.norm(direction)
                        if dist_mag > 0:
                            direction_norm = direction / dist_mag
                            cam.SetPosition(fp - direction_norm * target_dist)
                    cam.SetParallelProjection(False)
                else:
                    # Switching Perspective -> Parallel
                    dist = cam.GetDistance()
                    fov = cam.GetViewAngle()
                    new_scale = dist * np.tan(np.deg2rad(fov) / 2.0)
                    cam.SetParallelScale(new_scale)
                    cam.SetParallelProjection(True)
                
                renderer.ResetCameraClippingRange()
                render_window.Render()
            
            # Selection ('l')
            elif key_sym.lower() == 'l':
                click_pos = interactor.GetEventPosition()
                picker.Pick(click_pos[0], click_pos[1], 0, renderer)
                picked_actor = picker.GetActor()
                if picked_actor:
                    if picked_actor in selected_actors:
                        # Deselect
                        original_color = selected_actors.pop(picked_actor)
                        picked_actor.GetProperty().SetColor(*original_color)
                    else:
                        # Select
                        current_color = picked_actor.GetProperty().GetColor()
                        selected_actors[picked_actor] = current_color
                        picked_actor.GetProperty().SetColor(0.0, 1.0, 0.0) # Green
                render_window.Render()
            
            # Center ('c')
            elif key_sym.lower() == 'c':
                if selected_actors:
                    # Calculate center of all selected actors
                    sel_bounds = [inf, -inf, inf, -inf, inf, -inf]
                    for actor in selected_actors:
                        b = actor.GetBounds()
                        sel_bounds[0] = min(sel_bounds[0], b[0])
                        sel_bounds[1] = max(sel_bounds[1], b[1])
                        sel_bounds[2] = min(sel_bounds[2], b[2])
                        sel_bounds[3] = max(sel_bounds[3], b[3])
                        sel_bounds[4] = min(sel_bounds[4], b[4])
                        sel_bounds[5] = max(sel_bounds[5], b[5])
                    
                    center_x = (sel_bounds[0] + sel_bounds[1]) / 2.0
                    center_y = (sel_bounds[2] + sel_bounds[3]) / 2.0
                    center_z = (sel_bounds[4] + sel_bounds[5]) / 2.0
                    
                    # Move Focal Point
                    renderer.GetActiveCamera().SetFocalPoint(center_x, center_y, center_z)
                    renderer.ResetCameraClippingRange()
                    render_window.Render()

            # Hide selected ('h')
            elif key_sym.lower() == 'h':
                # Create a reverse lookup to find surface names from actor objects
                actor_to_name = {v: k for k, v in self.actors.items()}

                for actor in list(selected_actors.keys()):
                    # Hide the surface
                    original_color = selected_actors.pop(actor)
                    actor.GetProperty().SetColor(*original_color)
                    actor.SetVisibility(False)
                    
                    # Find and hide associated hits
                    if actor in actor_to_name:
                        
                        surface_name = actor_to_name[actor]
                        
                        # Look for keys starting with "hits_" + surface_name
                        target_prefix = f"hits_{surface_name}"
                        
                        for key, hit_actor in self.actors.items():
                            if key == target_prefix or key.startswith(target_prefix + "_"):
                                hit_actor.SetVisibility(False)

                render_window.Render()

            # Isolate selected ('i')
            elif key_sym.lower() == 'i':
                nonlocal is_isolated
                
                # If we have a selection, Isolate it and then Deselect
                if selected_actors:
                    is_isolated = True
                    # Hide everything not in selection
                    for name, actor in self.actors.items():
                        # Always keep rays visible
                        if "rays_" in name:
                            actor.SetVisibility(True)
                            continue
                        
                        if actor in selected_actors:
                            actor.SetVisibility(True)
                            # Restore color immediately
                            orig_col = selected_actors[actor]
                            actor.GetProperty().SetColor(*orig_col)
                        else:
                            actor.SetVisibility(False)
                    
                    # Clear selection since we are now in "Isolation Mode" with no active targets
                    selected_actors.clear()
                    
                # If nothing selected, 'i' acts as "Show All" if currently isolated
                elif is_isolated:
                    is_isolated = False
                    for actor in self.actors.values():
                        actor.SetVisibility(True)
                
                render_window.Render()

                # Consume the key, otherwise the axes will disappear (IDKW)
                interactor.SetKeySym("") 
                interactor.SetKeyCode("\0")

            # Clear selections ('Escape')
            elif key_sym == 'Escape':
                is_isolated = False
                # Restore colors of selected
                for actor, color in selected_actors.items():
                    actor.GetProperty().SetColor(*color)
                selected_actors.clear()
                # Ensure everything is visible
                for actor in self.actors.values():
                    actor.SetVisibility(True)
                render_window.Render()

            # Axes and modes
            elif key_sym.lower() == 'y':
                perform_view_reset((0,1,0))
            elif key_sym.lower() == 'x':
                perform_view_reset((1,0,0))
            elif key_sym.lower() == 'z':
                perform_view_reset((0,0,1))
            elif key_sym.lower() == 'w':
                for actor in renderer.GetActors():
                    actor.GetProperty().SetRepresentationToWireframe()
                render_window.Render()
            elif key_sym.lower() == 's':
                for actor in renderer.GetActors():
                    actor.GetProperty().SetRepresentationToSurface()
                render_window.Render()
            elif key_sym.lower() == 'p':
                for actor in renderer.GetActors():
                    actor.GetProperty().SetRepresentationToPoints()
                render_window.Render()
            elif key_sym.lower() == 'q':
                    quit(render_window_interactor)

        priority = 1.0
        render_window_interactor.AddObserver(vtk.vtkCommand.CharEvent, key_press_callback, priority)

        # Start
        render_window.Render()
        
        # Camera configuration
        cam = renderer.GetActiveCamera()
        
        # Apply initial projection type
        cam.SetParallelProjection(orthographic)

        if camera_position is not None:
            cam.SetPosition(*camera_position)
            cam.SetViewUp(*view_up)
            
            if focal_point is not None:
                if isinstance(focal_point, str):
                    fp = self.os[focal_point].position
                    if self._tel_to_local_rot is not None:
                        fp = self._tel_to_local_rot @ fp + self._translation_vector
                    cam.SetFocalPoint(*fp)
                else:
                    cam.SetFocalPoint(*focal_point)
            elif has_system_actors:
                cx = (sys_bounds[0] + sys_bounds[1]) / 2.0
                cy = (sys_bounds[2] + sys_bounds[3]) / 2.0
                cz = (sys_bounds[4] + sys_bounds[5]) / 2.0
                cam.SetFocalPoint(cx, cy, cz)
            else:
                cam.SetFocalPoint(0, 0, 0)
            
            renderer.ResetCameraClippingRange()
        else:
            if has_system_actors:
                renderer.ResetCamera(sys_bounds)
            else:
                renderer.ResetCamera()
        
        # Increase far clip plane
        rng = cam.GetClippingRange()
        cam.SetClippingRange(rng[0], rng[1] * 1000) 

        render_window.Render()
        render_window_interactor.Initialize()
        render_window_interactor.Start()

        # Stop
        quit(render_window_interactor)

    def _create_wavelength_lut(self):
        """Creates a color transfer function for wavelength in the range 200nm - 1000nm."""
        ctf = vtk.vtkColorTransferFunction()
        ctf.SetColorSpaceToRGB()
        
        # Bright white/blue
        ctf.AddRGBPoint(200.0, 0.9, 0.9, 1.0) 
        # Electric purple
        ctf.AddRGBPoint(300.0, 0.7, 0.5, 1.0)   
        
        # Standard Rainbow, fully saturated
        ctf.AddRGBPoint(380.0, 0.5, 0.0, 1.0)   # Violet
        ctf.AddRGBPoint(440.0, 0.0, 0.2, 1.0)   # Blue
        ctf.AddRGBPoint(490.0, 0.0, 1.0, 1.0)   # Cyan
        ctf.AddRGBPoint(510.0, 0.2, 1.0, 0.2)   # Green
        ctf.AddRGBPoint(580.0, 1.0, 1.0, 0.0)   # Yellow
        ctf.AddRGBPoint(600.0, 1.0, 0.5, 0.0)   # Orange
        ctf.AddRGBPoint(700.0, 1.0, 0.0, 0.0)   # Red
        
        # Bright cherry red
        ctf.AddRGBPoint(850.0, 0.9, 0.2, 0.2)   
        # Pale pink
        ctf.AddRGBPoint(1000.0, 0.7, 0.5, 0.5)
        
        return ctf

    def add_rays(self, start, stop, surface_id=None, wavelengths=None, directions=None, length=None, point_size=1.0, show_rays=True, show_hits=True):
        """
        Draw rays from start to stop positions and highlight stop points.
        
        If a stop position is NaN (indicating a miss), the ray is skipped by default.
        If 'length' and 'directions' are provided, rays with NaN stops are drawn 
        starting from 'start' along 'direction' for 'length' units (without a hit dot).
        
        Parameters
        ----------
        start : ndarray
            (n, 3) array of starting coordinates (x, y, z).
        stop : ndarray
            (n, 3) array of stopping coordinates (x, y, z).
        surface_id: ndarray
            (n,) array of surface indices reached by each ray.
        directions : ndarray, optional
            (n, 3) array of direction vectors. Required if plotting NaN rays with fixed length.
        length : float, optional
            If provided, rays with NaN stop will be drawn with this length using the direction vector.
        opacity : float, optional
            Transparency of the rays from 0.0 (invisible) to 1.0 (opaque). Default is 0.5.
        point_size : float, optional
            Size of the hit points. Default is 1.0
        show_rays : bool, optional
            Whether to show rays or not. Default is True.
        show_hits : bool, optional
            Whether to show hits or not. Default is True.
        """

        if not show_hits and not show_rays:
            return

        has_directions = directions is not None

        has_wavelengths = wavelengths is not None

        if start.shape != stop.shape:
            raise ValueError("Start and stop points must have the same shape.")

        if has_directions and start.shape != directions.shape:
            raise ValueError("Start/stop poinnts and directions must have the same shape.")
        
        if has_wavelengths and start.shape[0] != wavelengths.shape[0]:
            print(start.shape[0], wavelengths.shape[0])
            raise ValueError("Start/stop points and wavelengths shape mismatch.")
        
        wavelength_scalars = vtk.vtkFloatArray()
        wavelength_scalars.SetName("Wavelengths")
        
        n_photons = start.shape[0]
        
        # Rays
        points_rays = vtk.vtkPoints()
        lines_rays = vtk.vtkCellArray()
        
        # Hits
        hits_collections = {}
        if surface_id is None:
            surface_id = np.zeros((n_photons,), dtype=np.int32)

        # Check which rays have intersected a surface
        valid_stops = ~np.any(np.isnan(stop), axis=1)
        
        for i in range(n_photons):
            current_start = start[i]
            current_stop = stop[i]
            label = self.os[surface_id[i]].name
            
            is_valid_stop = valid_stops[i]
            
            final_stop = None
            draw_hit = False
            if is_valid_stop:
                # Draw from start to stop and add a hit
                final_stop = current_stop
                draw_hit = True
            elif length is not None and has_directions:
                # Draw fixed length ray if directions are available
                final_stop = current_start + directions[i] * length
                draw_hit = False
            else:
                # Ignore not intersected ray
                continue
            
            if show_rays:
                ## Ray logic
                id_start = points_rays.InsertNextPoint(current_start)
                id_stop_line = points_rays.InsertNextPoint(final_stop)
                
                line = vtk.vtkLine()
                line.GetPointIds().SetId(0, id_start)
                line.GetPointIds().SetId(1, id_stop_line)
                lines_rays.InsertNextCell(line)

                # Store wavelength
                if has_wavelengths:
                    wavelength_scalars.InsertNextValue(wavelengths[i])
            
            ## Hit logic
            if draw_hit:
                if label not in hits_collections:
                    hits_collections[label] = {
                        'points': vtk.vtkPoints(),
                        'verts': vtk.vtkCellArray()
                    }
                # Add point to specific surface collection
                pt_id = hits_collections[label]['points'].InsertNextPoint(final_stop)
                hits_collections[label]['verts'].InsertNextCell(1)
                hits_collections[label]['verts'].InsertCellPoint(pt_id)
            
        ## Rays actor
        if show_rays:
            raysPolyData = vtk.vtkPolyData()
            raysPolyData.SetPoints(points_rays)
            raysPolyData.SetLines(lines_rays)
            
            mapper_rays = vtk.vtkPolyDataMapper()
            mapper_rays.SetInputData(raysPolyData)

            if has_wavelengths:
                raysPolyData.GetCellData().SetScalars(wavelength_scalars)                
                # Use the lookup table
                lut = self._create_wavelength_lut()
                mapper_rays.SetLookupTable(lut)
                mapper_rays.SetScalarRange(200, 1000) # nm
                mapper_rays.SetScalarModeToUseCellData()
            else:
                mapper_rays.ScalarVisibilityOff()
            
            actor_rays = vtk.vtkActor()
            actor_rays.SetMapper(mapper_rays)

            if not has_wavelengths:
                actor_rays.GetProperty().SetColor(0.0, 0.7, 1.0) # Electric blue
            
            actor_rays.GetProperty().SetLineWidth(1.0)
            if self._ray_opacity is not None:
                actor_rays.GetProperty().SetOpacity(float(self._ray_opacity))
            
            # Make rays unpickable
            actor_rays.PickableOff()

            base_key = 'rays'
            key = base_key
            count = 0
            while key in self.actors:
                key = f"{base_key}_{count}"
                count += 1
            
            self.actors[key] = actor_rays

        ## Hits actor
        if show_hits:
            for label, data in hits_collections.items():
                if data['points'].GetNumberOfPoints() == 0:
                    continue
            
                dotsPolyData = vtk.vtkPolyData()
                dotsPolyData.SetPoints(data['points'])
                dotsPolyData.SetVerts(data['verts'])
                
                mapper_dots = vtk.vtkPolyDataMapper()
                mapper_dots.SetInputData(dotsPolyData)
                
                actor_dots = vtk.vtkActor()
                actor_dots.SetMapper(mapper_dots)
                actor_dots.GetProperty().SetColor(1.0, 1.0, 0.0)
                actor_dots.GetProperty().SetPointSize(point_size)
                actor_dots.GetProperty().SetOpacity(1.0)
                actor_dots.PickableOff()

                # Naming convention: hits_SurfaceID
                # If multiple batches are added, append an index
                base_key = f"hits_{label}"
                key = base_key
                count = 0
                while key in self.actors:
                    key = f"{base_key}_{count}"
                    count += 1
                
                self.actors[key] = actor_dots

    def _set_ray_opacity(self, value):
        """Set the opacity of rays.
        
        Parameters
        ----------
        value: float
            Ray opacity.
        
        """
        if value is None:
            return
        
        for actor_name in self.actors:
            if actor_name.startswith("rays"):
                self.actors[actor_name].GetProperty().SetOpacity(float(value))

    def set_ray_opacity(self, value):
        """Set the opacity of rays.
        
        Parameters
        ----------
        value: float
            Ray opacity.
        
        """
        self._ray_opacity = value
        self._set_ray_opacity(value)

    def _apply_global_transform(self):
        """
        Apply a global rotation and translation to all current actors.
        
        Parameters
        ----------
        R : ndarray
            (3, 3) Rotation matrix. 
        t : ndarray or list
            (3,) Translation vector (x, y, z).
        """
        # Convert numpy array to vtkMatrix4x4
        vtk_R = vtk.vtkMatrix4x4()
        for i in range(3):
            for j in range(3):
                vtk_R.SetElement(i, j, self._tel_to_local_rot[i, j])

        # Iterate over all actors
        for actor in self.actors.values():
            
            # Get the existing transform
            transform = actor.GetUserTransform()
            if transform is None:
                transform = vtk.vtkTransform()
                transform.SetMatrix(actor.GetMatrix())
                actor.SetUserTransform(transform)
            
            # Apply transformations
            # PostMultiply ensures we are applying this to the "Global" world coordinates
            transform.PostMultiply()
            
            # Apply rotation
            transform.Concatenate(vtk_R)
            
            # Apply translation
            transform.Translate(*self._translation_vector)

            transform.Update()

    def save_screenshot(self,
        filename,
        camera_position,
        focal_point=None,
        view_up=(0, 0, 1),
        size=None,
        image_scale=1,
        orthographic=False,
        representation='surface',
        hide_surfaces=None,
        zoom=1.0
    ) -> None:
        """
        Render the scene off-screen and save it to a PNG file.

        Coordinates for camera position, focal point, and view up should be provided 
        in the telescope coordinate system. The method handles the transformation 
        to the local coordinate system automatically.

        Parameters
        ----------
        filename : str
            Path to save the resulting PNG file.
        camera_position : sequence of float
            The (x, y, z) coordinates of the camera position.
        focal_point : str or sequence of float, optional
            The point the camera is looking at. 
            - If None (default): The camera looks at the center of the bounding box of all system actors.
            - If str: Looks for an optical surface with this name and targets its position.
            - If sequence: Uses the provided (x, y, z) coordinates.
        view_up : sequence of float
            The vector defining the "up" direction for the camera, default=(0, 0, 1).
        size : tuple of int, optional
            The (width, height) of the rendered image in pixels. 
            If None, uses the current window size.
        image_scale : int
            Resolution multiplier (e.g., 2 doubles the pixel density), default=1.
        orthographic : bool
            If True, uses parallel projection.
            If False, uses standard perspective projection.
            By default False.
        representation : {'surface', 'wireframe', 'points'}
            The rendering style of the actors in the scene, default='surface'.
        hide_surfaces : list of str, optional
            A list of names of surfaces/actors to exclude from the render.
        zoom : float
            Zoom factor applied when ``orthographic`` is True.
            - 1.0: Fits the object diagonal to the screen height.
            - > 1.0: Zooms in.
            - < 1.0: Zooms out.
            By default 1.0.
            Note: This parameter currently implies no effect if orthographic is False.

        """
        # Setup Renderer
        renderer = vtk.vtkRenderer()
        renderer.SetBackground(0.1, 0.15, 0.25)
        render_window = vtk.vtkRenderWindow()
        render_window.SetOffScreenRendering(1)
        render_window.AddRenderer(renderer)
        
        if size: render_window.SetSize(*size)
        else: render_window.SetSize(*self.window_size)

        # Ray Opacity
        if self._ray_opacity is None:
            n_rays = 0
            for actor_name, actor in self.actors.items():
                if actor_name.startswith('rays'):
                    n_rays += actor.GetMapper().GetDataSetInput().GetLines().GetNumberOfCells()
            self._set_ray_opacity(self._get_sigmoid_opacity(n_rays))

        # Add actors and calculate bounds
        inf = float('inf')
        sys_bounds = [inf, -inf, inf, -inf, inf, -inf]
        has_system_actors = False
        hidden_set = set(hide_surfaces) if hide_surfaces else set()

        for name, actor in self.actors.items():
            # Check hidden
            should_skip = name in hidden_set
            if not should_skip and name.startswith("hits_"):
                for hidden_name in hidden_set:
                    if name.startswith(f"hits_{hidden_name}"):
                        should_skip = True; break
            if should_skip: continue

            # Set representation
            prop = actor.GetProperty()
            if representation == 'wireframe': prop.SetRepresentationToWireframe()
            elif representation == 'points': prop.SetRepresentationToPoints()
            else: prop.SetRepresentationToSurface()

            renderer.AddActor(actor)
            
            # Update bounds (using world coordinates of the actor)
            if not name.startswith("rays"):
                has_system_actors = True
                b = actor.GetBounds()
                sys_bounds[0] = min(sys_bounds[0], b[0])
                sys_bounds[1] = max(sys_bounds[1], b[1])
                sys_bounds[2] = min(sys_bounds[2], b[2])
                sys_bounds[3] = max(sys_bounds[3], b[3])
                sys_bounds[4] = min(sys_bounds[4], b[4])
                sys_bounds[5] = max(sys_bounds[5], b[5])

        # Configure camera coordinates
        cam = renderer.GetActiveCamera()
        cam.SetParallelProjection(orthographic)

        # Transform inputs to world Coordinates
        final_pos = np.array(camera_position, dtype=float)
        final_up = np.array(view_up, dtype=float)
        if self._tel_to_local_rot is not None:
            final_pos = self._tel_to_local_rot @ final_pos + self._translation_vector
            final_up = self._tel_to_local_rot @ final_up

        cam.SetPosition(*final_pos)
        cam.SetViewUp(*final_up)

        # Handle focal point
        final_fp = np.zeros(3)
        if focal_point is not None:
            if isinstance(focal_point, str):
                # Look for surface by name
                found = False
                for s in self.os:
                    if s.name == focal_point:
                        fp_local = np.array(s.position, dtype=float)
                        if self._tel_to_local_rot is not None:
                            final_fp = self._tel_to_local_rot @ fp_local + self._translation_vector
                        else:
                            final_fp = fp_local
                        found = True
                        break
                if not found: final_fp = np.array([0., 0., 0.])
            else:
                fp_input = np.array(focal_point, dtype=float)
                if self._tel_to_local_rot is not None:
                    final_fp = self._tel_to_local_rot @ fp_input + self._translation_vector
                else:
                    final_fp = fp_input
        elif has_system_actors:
            # Center of bounding box
            final_fp[0] = (sys_bounds[0] + sys_bounds[1]) / 2.0
            final_fp[1] = (sys_bounds[2] + sys_bounds[3]) / 2.0
            final_fp[2] = (sys_bounds[4] + sys_bounds[5]) / 2.0

        cam.SetFocalPoint(*final_fp)

        # Handle orthographic scaling
        if orthographic and has_system_actors:
            # Calculate diagonal length of the bounding box
            dx = sys_bounds[1] - sys_bounds[0]
            dy = sys_bounds[3] - sys_bounds[2]
            dz = sys_bounds[5] - sys_bounds[4]
            diagonal = np.sqrt(dx*dx + dy*dy + dz*dz)
            
            # Set scale so the whole object fits, adjusted by zoom
            # ParallelScale is half the height of the viewport
            if zoom == 0:
                zoom = 1.0
            cam.SetParallelScale((diagonal / 2.0) / zoom)

        renderer.ResetCameraClippingRange()
        render_window.Render()

        # Capture
        w2if = vtk.vtkWindowToImageFilter()
        w2if.SetInput(render_window)
        w2if.SetScale(image_scale)
        w2if.SetInputBufferTypeToRGB()
        w2if.ReadFrontBufferOff()
        w2if.Update()

        writer = vtk.vtkPNGWriter()
        writer.SetFileName(filename)
        writer.SetInputConnection(w2if.GetOutputPort())
        writer.Write()
        
        print(f"Screenshot saved to {filename}")


import os
import sys
from pathlib import Path

def _configure_conda_env():
    if len(sys.argv) < 2:
        print("Usage: python setup_env_scripts.py <target_env_name>", file=sys.stderr)
        sys.exit(1)
    target_env = sys.argv[1]

    conda_prefix_str = os.getenv("CONDA_PREFIX")
    if not conda_prefix_str:
        print("Error: CONDA_PREFIX environment variable not set.", file=sys.stderr)
        print("Ensure you are running this script within an active Conda environment.", file=sys.stderr)
        sys.exit(1)

    conda_prefix = Path(conda_prefix_str)
    current_env = conda_prefix.name

    if current_env != target_env:
        print(f"Error: The target environment '{target_env}' is not active.", file=sys.stderr)
        print(f"Currently active environment: '{current_env}'", file=sys.stderr)
        sys.exit(1)

    # Define activation and deactivation directories
    activate_dir = conda_prefix / "etc" / "conda" / "activate.d"
    deactivate_dir = conda_prefix / "etc" / "conda" / "deactivate.d"

    activate_dir.mkdir(parents=True, exist_ok=True)
    deactivate_dir.mkdir(parents=True, exist_ok=True)

    # NVHPC_ROOT is assumed to be defined when `module load nvhpc` runs
    # CUDA_PATH is assumed to be $NVHPC_ROOT/cuda
    set_vars_script = """\
#!/bin/sh

module load nvhpc

if [ -n "$NVHPC_ROOT" ]; then
  export NVCC="$NVHPC_ROOT/compilers/bin/nvcc"
  export CUDA_PATH="$NVHPC_ROOT/cuda"
else
  echo "Warning: NVHPC_ROOT was not set after 'module load nvhpc'." >&2
fi
"""

    unset_vars_script = """\
#!/bin/sh

unset NVCC
unset CUDA_PATH
module unload nvhpc
"""

    set_vars_path = activate_dir / "set_vars.sh"
    try:
        with open(set_vars_path, "w") as f:
            f.write(set_vars_script)
        os.chmod(set_vars_path, 0o755)
    except IOError as e:
        print(f"Error writing activation script: {e}", file=sys.stderr)
        sys.exit(1)

    unset_vars_path = deactivate_dir / "unset_vars.sh"
    try:
        with open(unset_vars_path, "w") as f:
            f.write(unset_vars_script)
        os.chmod(unset_vars_path, 0o755)
    except IOError as e:
        print(f"Error writing deactivation script: {e}", file=sys.stderr)
        sys.exit(1)


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

from typing import List, Dict, Optional
from tabulate import tabulate

from ._surface import AsphericalSurface, Surface
from ._surface_misc import SurfaceType

# TODO: add insert e ramove methods
class OpticalSystem():
    """Represents a generic optical system.

    Parameters
    ----------
    name : str, optional
        Name of the optical system, by defaults "OS".
    surfaces : List[Surface]
        List of Surface objects representing the optical surfaces in the
        system. The order of the surfaces in the list defines the sequential
        order for ray tracing.

    Notes
    -----
    The order of surfaces for ray tracing is determined by the order of the
    objects in the `surfaces` list.
    Each surface has a `name` attribute that is used for named access. If a
    surface's name is None, a default name "Surface_i" (where i is the index)
    is assigned.

    """

    def __init__(
        self,
        surfaces: List[Surface],
        name: Optional[str] = None,
    ):
        self._name = name if name is not None else "OS"
        self._surfaces: List[Surface] = []
        self._surface_map: Dict[str, Surface] = {}

        for i, surface in enumerate(surfaces):
            if surface.name is None:
                default_name = f"Surface_{i}"
                surface.name = default_name  # Assign a name to the surface.
            else:
                default_name = surface.name

            if default_name in self._surface_map:
                raise ValueError(f"Duplicate surface name: {default_name}")

            self._surfaces.append(surface)
            self._surface_map[default_name] = surface
        
        if not any([self._is_surface_sensitive(s) for s in self]):
            raise(RuntimeError(f'No sensitive surfaces defined for OpticalSystem {self._name}.'))

    @staticmethod
    def _is_surface_sensitive(surface):
        needed = [
            SurfaceType.SENSITIVE,
            SurfaceType.SENSITIVE_BACK,
            SurfaceType.SENSITIVE_FRONT,
            SurfaceType.REFLECTIVE_SENSITIVE,
        ]
        return any([surface.type == t for t in needed])

    @property
    def name(self) -> str:
        return self._name

    def remove_surface(self, surface):
        """Remove a given surface.

        Parameters
        ----------
        surface : Surface or str
            Surface instance or name to be removed.
        """
        if isinstance(surface, str):
            if surface not in self:
                return
            surface = self[surface]
        
        self._surface_map.pop(surface.name)
        self._surfaces = list(self._surface_map.values())


    def add_surface(self, surface, replace=False):
        """Add a surface to the optical syste.

        Parameters
        ----------
        surface : Surface
            Surface to be added.
        replace : bool, optional
            Whether to replace an exisisting surface with the same name, by default False.

        Raises
        ------
        ValueError
            If the surface already exists and `replace` is `False`.
        """
        if surface in self:
            if not replace:
                raise ValueError(f"Surface '{surface.name}' already placed. If you want to overwrite it use replace=True.")
            self.remove_surface(surface)
        
        if surface.name is None:
            surface.name = f"Surface_{len(self)}"

        self._surfaces.append(surface)
        self._surface_map[surface.name] = surface

    @property
    def surfaces_dict(self) -> Dict[str, AsphericalSurface]:
        """Provides a dictionary view of the surfaces, mapping names to :py:class:`Surface` objects."""
        return self._surface_map.copy()

    def __contains__(self, x):
        if issubclass(type(x), Surface):
            x = x.name
        if x in self._surface_map.keys():
            return True

    def __iter__(self):
        """
        Returns an iterator over the optical surfaces.

        Returns
        -------
        iterator
            An iterator object that iterates through the surfaces in `self._surfaces`.
        """
        return iter(self._surfaces)

    def __len__(self):
        """
        Returns the number of optical surfaces in the system.

        Returns
        -------
        int
            The number of surfaces.
        """
        return len(self._surfaces)

    def __getitem__(self, key: str | int):
        """
        Allows accessing individual surfaces using indexing or by surface name.

        Parameters
        ----------
        key : str | int
            The name or index of the desired surface.

        Returns
        -------
        OpticalSurface
            The optical surface with the specified name or at the specified index.

        Raises
        ------
        KeyError
            If a surface with the given name is not found.
        IndexError
            If the index is out of range.
        TypeError
            If the key is neither a string nor an integer.
        """
        if isinstance(key, str):
            if key not in self._surface_map:
                raise KeyError(f"No surface found with name '{key}'")
            return self._surface_map[key]
        elif isinstance(key, int):
            return self._surfaces[key]
        else:
            return self._surfaces[int(key)]

    def _get_table_data(self, tablefmt: str = "grid") -> str:
        """Constructs the data for tabulate optical system info.

        Parameters
        ----------
        tablefmt : str
            Format specifier for tabulate, by defaults "grid"

        Returns
        -------
        str
            Tabulated string representation of the optical system data
        """
        data = []
        for name, surface in self.surfaces_dict.items():
            material_in_name = (
                surface.material_in.name
                if surface.material_in is not None
                else "None"
            )
            material_out_name = (
                surface.material_out.name
                if surface.material_out is not None
                else "None"
            )
            data.append([name, f"{surface.position[2]:3.3f}", material_in_name, material_out_name, surface.type.name])

        headers = ["Surface Name", "Axial Position", "Material above", "Material below", "Surface Type"]

        return tabulate(data, headers=headers, tablefmt=tablefmt)

    def __repr__(self) -> str:
        """Returns a string representation of the optical system."""
        title = f"**Optical System: {self.name}**\n"
        return title + self._get_table_data()

    def _repr_html_(self) -> str:
        """Returns an HTML representation of the optical system for Jupyter Notebook."""
        title = f"<h5>Optical System: {self.name}</h5>"
        table = self._get_table_data(tablefmt="html")
        return title + table

    def _repr_markdown_(self) -> str:
        """Returns a Markdown representation of the optical system."""
        title = f"**Optical System: {self.name}**\n\n"
        table = self._get_table_data(tablefmt="pipe")
        return title + table


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np

def normalize_aspheric_coefficients(coefficients: np.ndarray | list | tuple, half_aperture: float) -> np.ndarray:
    """Rewrite aspheric coefficients for float32 calculations.

    This function normalizes aspheric coefficients to ensure numerical stability
    when using single-precision calculations. It scales the coefficients
    based on the surface aperture radius to prevent potential overflow or
    underflow issues during calculations involving powers of the radial distance.

    Parameters
    ----------
    coefficients : numpy.ndarray | list | tuple
        The aspheric coefficients. It can be a NumPy array, a list or a tuple.
    half_aperture : float
        The half-aperture of the surface.

    Returns
    -------
    new_coefficients : numpy.ndarray
        A new array containing the normalized aspheric coefficients.

    Notes
    -----
    The normalization is done by multiplying each coefficient A[i] by ra^(2*(i+1)).
    This scaling ensures that the coefficients are adjusted appropriately for
    calculations involving the radial distance, which is typically normalized by
    the surface half-aperture.
    """
    new_coefficients = np.array(coefficients, copy=True)  # Convert to NumPy array if it is a list or tuple
    for i in range(new_coefficients.shape[0]):
        new_coefficients[i] = new_coefficients[i] * half_aperture**(2*(i+1))
    return new_coefficients


import os
import yaml

class PathLoader(yaml.SafeLoader):
    """
    Custom YAML loader that resolves file paths relative to the configuration file location.

    This loader enables the use of the ``!path`` tag in YAML files. When this tag is encountered,
    the loader automatically resolves the specified path relative to the directory containing
    the YAML file, rather than the current working directory of the Python script.

    Parameters
    ----------
    stream : file object
        The file stream to load. It must have a ``name`` attribute containing
        the file path (e.g., a file object created via ``open()``).

    """

    def __init__(self, stream):
        # Determine the folder of the file being loaded
        self._root = os.path.split(stream.name)[0]
        super().__init__(stream)

    def construct_path(self, node):
        """
        Construct a path by joining the file root directory with the node value.

        Parameters
        ----------
        node : yaml.nodes.ScalarNode
            The YAML node containing the relative path string.

        Returns
        -------
        str
            The absolute or relative path resolved against the configuration file directory.
        """
        # This function runs whenever the loader sees '!path'
        value = self.construct_scalar(node)
        return os.path.normpath(os.path.join(self._root, value))

# Register the constructor
PathLoader.add_constructor('!path', PathLoader.construct_path)


# Copyright (C) 2024- Davide Mollica <davide.mollica@inaf.it>
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of iactsim.
#
# iactsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# iactsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iactsim.  If not, see <https://www.gnu.org/licenses/>.

from pathlib import Path
from typing import Tuple, Union, Dict, Any

import numpy as np
import yaml

from ...optics._surface_misc import SurfaceType
from ...optics._surface import (
    AsphericalSurface,
    FlatSurface,
    SphericalSurface,
    CylindricalSurface,
    ApertureShape,
    SurfaceProperties,
    SipmTileSurface
)
from ...optics._optical_system import OpticalSystem
from ...optics._materials import Materials
from ...optics._cpu_transforms import photon_to_local_rotation

from ._camera import AstriCameraGeometry

from ...io._path_loader import PathLoader

class AstriOpticalSystem(OpticalSystem):
    """ASTRI optical system class.

    See Also
    --------
        :py:class:`iactsim.optics.OpticalSystem`:
    """
    def __init__(self):
        #####
        # M1
        #####
        r_1 = 2153.
        c_1 = 1/8223.
        k_1 = 0.
        A_1 = -np.array([
            0.0,
            9.610594335657154e-13,
            -5.655007251743067e-20,
            6.779846300812014e-27,
            3.895878871219899e-33,
            5.280381799628059e-40,
            -2.991060741547783e-47,
            -4.391530590471279e-53,
            -6.174332023811481e-60,
            2.7358654573036966e-66
        ])
        self.m1 = AsphericalSurface(
            r_1, c_1, k_1, A_1,
            surface_type=SurfaceType.REFLECTIVE_FRONT,
            aperture_shape=ApertureShape.CIRCULAR,
            central_hole_shape=ApertureShape.HEXAGONAL,
            central_hole_half_aperture=423.,
            name='M1'
        )

        #####
        # M2
        #####
        r_2 = 900.
        c_2 = -1/2180.
        k_2 = 0.
        A_2 = -np.array([
            0.0,
            1.620759030635574e-11,
            -2.895843564948884e-17,
            8.633712193781264e-24,
            3.348558057289531e-30,
            -1.033610024069273e-36,
            -6.735253655133213e-43,
            -3.0654711881577763e-49,
            3.1716123568176433e-55,
            -3.711831037954923e-62
        ])
        self.m2 = AsphericalSurface(
            r_2, c_2, k_2, A_2,
            surface_type=SurfaceType.REFLECTIVE_BACK,
            name='M2'
        )

        ###############
        # Focal Plane
        ###############
        c_fp = -1./1060.
        r_fp = 250.
        self.fp = SphericalSurface(
            half_aperture=r_fp,
            curvature=c_fp,
            surface_type=SurfaceType.SENSITIVE_FRONT,
            name='FP'
        )

        # M1 at origin
        self.m1.position = (0, 0, 0)

        # M1-M2 distance
        self._m1_m2_distance = 3108.4
        self.m2.position = (0, 0, self._m1_m2_distance)

        # M2-FP distance
        self._back_focal_length = 519.6
        self.fp.position = (0, 0, self.m2.axial_position - self._back_focal_length)

        #######################
        # Dummy camera window
        #######################
        self._camera_window_radius = 250.
        self._dummy_window = FlatSurface(
            self._camera_window_radius,
            position=(0,0,self.m2.axial_position-self._back_focal_length+1),
            surface_type=SurfaceType.DUMMY,
            name=f'Window'
        )

        ##############
        # Camera body
        ##############
        self._camera_body_size = [250, 400]
        camera_body = CylindricalSurface(
            radius=self._camera_body_size[0],
            height=self._camera_body_size[1],
            surface_type=SurfaceType.OPAQUE,
            name='CameraBody'
        )
        # Top of the body on the window surface closer to M2
        camera_body.position = (0., 0., self._dummy_window.axial_position-0.5*self._camera_body_size[1])
        camera_body.tilt_angles = (0., 0., 0.)
        camera_body.top = False
        camera_body.bottom = True

        ###############
        # Central tube
        ###############
        central_tube_height = camera_body.position[2]-self._camera_body_size[1]*0.5
        central_tube = CylindricalSurface(
            radius=203.2,
            height=central_tube_height,
            position=(0,0,0.5*central_tube_height),
            has_bottom=True,
            has_top=True,
            surface_type=SurfaceType.OPAQUE,
            name='CentralTube'
        )

        #################################################
        # M2 baffle (should be updated if M2 is tilted)
        #################################################
        m2_baffle_pos = self.m2.position.copy()
        m2_baffle_pos[2] += self.m2.sagitta(900)
        m2_baffle_surface_1 = FlatSurface(
            position = m2_baffle_pos,
            half_aperture = 1070.,
            central_hole_half_aperture=900.,
            surface_type=SurfaceType.OPAQUE,
            name='M2-baffle-1'
        )

        baffle_dz = 150.4 # 3108.4-2958
        m2_baffle_surface_2 = CylindricalSurface(
            position = m2_baffle_pos-np.asarray([0,0,0.5*baffle_dz]),
            radius = 1070.,
            height = baffle_dz,
            surface_type=SurfaceType.OPAQUE,
            has_bottom=False,
            has_top=False,
            name='M2-baffle-2'
        )

        self.camera_geometry = AstriCameraGeometry()

        super().__init__(surfaces=[self.m1,self.m2,self._dummy_window,self.fp,camera_body,central_tube, m2_baffle_surface_1, m2_baffle_surface_2], name='ASTRI-OS')

    def build_masts(self):
        # ASTRI telescope have 3 group of masts radially simmetric
        # In the following the first group is defined and then 
        # the other two groups are generated from a rotation
        # of the first group around the optical axis.

        ##################################
        # Define the first group of masts
        ##################################

        # Main mast
        mast1_radius = 50.8
        mast1_height = 3135.28 + 100
        mast1_position = [0,-1729.66,1407.78]
        # Secondary masts
        mast2_radius = mast3_radius = 38.05
        mast2_height = 2290.78-50
        mast2_position = [0,-1214.62,1162.12]
        mast3_height = 1156.94
        mast3_position = [0,-705.21,2311.20]
        # Negligible
        hollow = True

        # Define telescope mast group (with a x-rotation)
        mast1_1 = CylindricalSurface(
            radius=mast1_radius,
            height=mast1_height,
            has_bottom=not hollow,
            has_top=not hollow,
            position=mast1_position,
            tilt_angles=((63.9-90),0,0),
            name="Mast1-1"
        )
        mast1_2 = CylindricalSurface(
            radius=mast2_radius,
            height=mast2_height,
            has_bottom=not hollow,
            has_top=not hollow,
            position=mast2_position,
            tilt_angles=((32.6-90),0,0),
            name="Mast1-2"
        )
        mast1_3 = CylindricalSurface(
            radius=mast3_radius,
            height=mast3_height,
            has_bottom=not hollow,
            has_top=not hollow,
            position=mast3_position,
            tilt_angles=((90-43.9),0,0),
            name="Mast1-3"
        )
        self.add_surface(mast1_1, replace=True)
        self.add_surface(mast1_2, replace=True)
        self.add_surface(mast1_3, replace=True)

        ####################################
        # Define the other mast groups 
        # rotating the first by 60 deg and 
        # 120 deg around the telescope axis
        ####################################
        theta = 2*np.pi/3
        for i in range(1,3):
            # z-rotation matrix
            cos_z = np.cos(i*theta)
            sin_z = np.sin(i*theta)
            z_rot = np.array([
                [ cos_z, -sin_z, 0.],
                [ sin_z,  cos_z, 0.],
                [    0.,     0., 1.],
            ])
            # New masts to be rotated
            mast_i_1 = CylindricalSurface(
                radius=mast1_radius,
                height=mast1_height,
                has_bottom=not hollow,
                has_top=not hollow,
                position=z_rot@mast1_1.position,
                name=f"Mast{i+1}-1"
            )
            mast_i_2 = CylindricalSurface(
                radius=mast2_radius,
                height=mast2_height,
                has_bottom=not hollow,
                has_top=not hollow,
                position=z_rot@mast1_2.position,
                name=f"Mast{i+1}-2"
            )
            mast_i_3 = CylindricalSurface(
                radius=mast3_radius,
                height=mast3_height,
                has_bottom=not hollow,
                has_top=not hollow,
                position=z_rot@mast1_3.position,
                name=f"Mast{i+1}-3"
            )

            # Define the first mast group then replicate the group with two consecutive 60 deg rotation around z-axis
            # rot = get_rotation_matrix().T ---> surface to telescope, the mast orientation in the telescope reference system
            # new_rot = z_rot @ rot ---> surface to telscope, rotated mast orientation in the telescope reference system
            # update_surface_rot = new_rot.T ---> telescope to surface!
            for mast_i_x, mast1_x in zip([mast_i_1,mast_i_2,mast_i_3],[mast1_1,mast1_2,mast1_3]):
                surface_to_telescope = mast1_x.get_rotation_matrix().T
                updated_surface_to_telescope = z_rot @ surface_to_telescope
                updated_telescope_to_surface = updated_surface_to_telescope.T
                mast_i_x.set_rotation_matrix(updated_telescope_to_surface)
            
            # Finally, add the rotated mast group
            self.add_surface(mast_i_1, replace=True)
            self.add_surface(mast_i_2, replace=True)
            self.add_surface(mast_i_3, replace=True)

    @property
    def camera_body_size(self):
        """Cylindrical camera body dimensions (radius and height)."""
        return self._camera_body_size
    
    def build_camera_modules(self):
        if not 'FP' in self:
            self.add_surface(self.fp, replace=True)
        
        fp = self['FP']
        
        cam_geom = self.camera_geometry
        
        pdm_p = cam_geom.modules_p
        pdm_n = cam_geom.modules_n
        for k in range(37):
            pdm = SipmTileSurface(
                pixels_per_side=8,
                pixel_active_side=cam_geom.pixel_active_side,
                pixels_separation=cam_geom.pixels_separation,
                border_to_active_area=0.2,
                surface_type=SurfaceType.REFLECTIVE_SENSITIVE,
                position = pdm_p[k] + fp.position,
                tilt_angles=(0.,0.,0.),
                name = f'PDM{k+1}'
            )
            pdm.sensor_id = k
            R = photon_to_local_rotation(pdm_n[k]).T
            pdm.set_rotation_matrix(R)
            self.add_surface(pdm, replace=True)

        self.remove_surface('FP')

    def remove_camera_modules(self):
        for k in range(37):
            s_name = f'PDM{k+1}'
            if s_name in self:
                self.remove_surface(s_name)

        self.add_surface(self.fp, replace=True)
    
    @camera_body_size.setter
    def camera_body_size(self, a_size):
        if 'Window' in self:
            last_surface = self['Window']
            p0 = last_surface.position
            fp_radius = last_surface.half_aperture
        elif 'WindowLayer3':
            last_surface = self['WindowLayer3']
            p0 = last_surface.position
            p0[2] += 0.5*last_surface.height
            fp_radius = last_surface.radius
        else:
            if 'FP' in self:
                last_surface = self['FP']
            else:
                last_surface = self['PDM19']
            p0 = last_surface['FP'].position
            fp_radius = self._camera_window_radius
        
        self._camera_body_size = a_size
        
        if 'CameraBody' in self:
            self['CameraBody'].radius = a_size[0]
            self['CameraBody'].height = a_size[1]
            self['CameraBody'].position = (0, 0, p0[2]-0.5*a_size[1]+17.3) # TODO:  edge between top filter and camera body

            if a_size[0]-fp_radius < 0:
                raise(ValueError("camera body size must be greater than camera window size."))
            
            if a_size[0]-fp_radius > 0.1: #mm
                top_camera_body = FlatSurface(
                    half_aperture = a_size[0],
                    central_hole_half_aperture=fp_radius,
                    surface_type=SurfaceType.OPAQUE,
                    position=(p0[0], p0[1], p0[2]+17.3), # TODO:  edge between top filter and camera body
                    name='CameraBodyTop'
                )
                self.add_surface(top_camera_body, replace=True)
        
            if 'CentralTube' in self:
                self['CentralTube'].height = self['CameraBody'].position[2]-self['CameraBody'].height*0.5
                self['CentralTube'].position = (0,0,0.5*self['CentralTube'].height)

    @property
    def camera_window_parameters(self):
        """Camera window parameterization:

            - window radius
            - SiPM window distance
            - layers thickness
            - layers separation
            - focus shift (opposite to the pointing direction)
        
        If None (or all None) a dummy flat surface is used.

        """
        return self._camera_window_parameters
    
    @camera_window_parameters.setter
    def camera_window_parameters(self, params):
        if hasattr(params, '__len__'):
            if all([x is None for x in params]):
                params = None
        
        if params is None:
            for k in range(3):
                self.remove_surface(f'WindowLayer{k+1}')
            self.remove_surface('Window')
            self.add_surface(self._dummy_window)
            self['CameraBody'].position = (0,0,self._dummy_window.axial_position-0.5*self._camera_body_size[1]+17.3) # TODO: edge between top filter and camera body
            if 'FP' in self:
                self['FP'].position = (0, 0, self['M2'].axial_position - self._back_focal_length)
            else:
                self.remove_camera_modules()
                self['FP'].position = (0, 0, self['M2'].axial_position - self._back_focal_length)
                self.build_camera_modules()
        else:
            # Remove dummy window
            self.remove_surface('Window')

            window_radius = params[0]
            sipm_window_distance = params[1]
            layers_thickness = params[2]
            layers_separation = params[3]
            focus_shift = params[4]

            fp_z_pos = self['M2'].axial_position - self._back_focal_length - focus_shift

            if 'FP' in self:
                self['FP'].position = (0, 0, fp_z_pos)
            else:
                self.remove_camera_modules()
                self['FP'].position = (0, 0, fp_z_pos)
                self.build_camera_modules()

            z = fp_z_pos + sipm_window_distance + 0.5*layers_thickness
            for i in range(3):
                layer = CylindricalSurface(
                    radius=window_radius,
                    height=layers_thickness,
                    name = f'WindowLayer{i+1}',
                    surface_type=SurfaceType.REFRACTIVE,
                    position=(0,0,z)
                )
                layer.material_in = Materials.FUSED_SILICA
                layer.material_out = Materials.AIR
                self.add_surface(layer, replace=True)

                z = z + layers_separation + layers_thickness

        # Update camera body position
        self.camera_body_size = self.camera_body_size

        self._camera_window_parameters = params

    def configure(self, config: Union[str, Path, Dict[str, Any]]) -> None:
        """Configure the optical system with a yaml configuration file or a dictionary.

        Parameters
        ----------
        config : str, path-like, or dict
            Path to the configuration file (str or pathlib.Path),
            or a dictionary containing the configuration.
        """
        # yaml file
        if isinstance(config, (str, Path)):
            with open(config, 'r') as file:
                os_conf = yaml.load(file, Loader=PathLoader)
        # Dictionary
        elif isinstance(config, dict):
            os_conf = config
        else:
            raise TypeError(
                "config must be a string (path), pathlib.Path, or a dictionary, "
                f"not {type(config)}"
            )

        if 'include_masts' in os_conf:
            if os_conf['include_masts'] == True:
                self.build_masts()

        if 'camera_window_parameters' in os_conf:
            values = os_conf['camera_window_parameters']
            if values is not None and len(values) < 5:
                raise(ValueError('Not enough values provided to configure the camera window (window radius, SiPM-window distance, layers thickness, layers separation and focus shift)'))
        else:
            values = None
        self.camera_window_parameters = values
        
        if 'camera_body_size' in os_conf:
            values = os_conf['camera_body_size']
            if len(values) < 1:
                raise(ValueError('Not enough values provided to configure the camera body (radius and depth.)'))
            self.camera_body_size = values

        for surface_name in ['M1', 'M2']:

            if surface_name not in os_conf:
                continue
            
            if 'M1' not in self:
                self.add_surface(self.m1)
            
            surface_dict = os_conf[surface_name]

            if 'properties' in surface_dict:
                prop_file = surface_dict['properties']
                surface_properties = SurfaceProperties.load_json(prop_file)
                self[surface_name].properties = surface_properties     
            
            scattering_dispersion = 0.
            if 'scattering_dispersion' in surface_dict:
                scattering_dispersion = surface_dict['scattering_dispersion']
            self[surface_name].scattering_dispersion = scattering_dispersion

        if 'window' in os_conf:
            
            surface_dict = os_conf['window']

            if "WindowLayer1" in self:
                layers = ["WindowLayer1", "WindowLayer2", "WindowLayer3"]
            else:
                layers = ["Window"]

            for s_name in layers:
                if 'layer_properties' in surface_dict:
                    prop_file = surface_dict['layer_properties']
                    surface_properties = SurfaceProperties.load_json(prop_file)
                    self[s_name].properties = surface_properties   
            
        self.build_camera_modules()
        
        if 'sipm' in os_conf:
            sipm_conf = os_conf['sipm']

            s_type = SurfaceType.SENSITIVE_FRONT
            if 'is_reflective' in sipm_conf:
                if sipm_conf['is_reflective']:
                    s_type = SurfaceType.REFLECTIVE_SENSITIVE

            surface_properties = None
            if 'surface_properties' in sipm_conf:
                prop_file = sipm_conf['surface_properties']
                surface_properties = SurfaceProperties.load_json(prop_file)
                if s_type == SurfaceType.REFLECTIVE_SENSITIVE:
                    surface_properties.efficiency_kind = 'absorbed'
                else:
                    surface_properties.efficiency_kind = 'incident'
                self.fp.properties = surface_properties
                self.fp.type = s_type

            for k in range(37):
                s_name = f'PDM{k+1}'
                if s_name in self:
                    self[s_name].properties = surface_properties
                    self[s_name].type = s_type

        if "M1-segments" in os_conf:
            segments_conf = os_conf["M1-segments"]

            use_segments = True
            if 'use' in segments_conf:
                use_segments = segments_conf['use']
            
            if use_segments:
                for key in segments_conf:
                    # Skip 'use' key
                    if key == 'use': 
                        continue

                    segment_conf = segments_conf[key] 
                    
                    offset = segment_conf['offset']
                    if offset is None:
                        raise(RuntimeError(f'Offset for segment {key} is not defined.'))

                    half_aperture = segment_conf['half_aperture']
                    if half_aperture is None:
                        raise(RuntimeError(f'Half aperture for segment {key} is not defined.'))

                    shift = [0,0,0]
                    if 'position_shift' in segment_conf:
                        position_shift = segment_conf['position_shift']
                        if position_shift is not None:
                            for i in range(len(position_shift)):
                                shift[i] = position_shift[i]
                    
                    tilt_angles = [0,0,0]
                    if 'tilt_angles' in segment_conf:
                        angles = segment_conf['tilt_angles']
                        if angles is not None:
                            for i in range(len(angles)):
                                tilt_angles[i] = angles[i]
                    
                    scattering_dispersion = None
                    if 'scattering_dispersion' in segment_conf:
                        scattering_dispersion = segment_conf['scattering_dispersion']
                    
                    if scattering_dispersion is None:
                        if 'M1' in self:
                            scattering_dispersion = self['M1'].scattering_dispersion
                        else:
                            scattering_dispersion = 0.
                    
                    curvature = None
                    if 'curvature' in segment_conf:
                        curvature = segment_conf['curvature']
                    if curvature is None:
                        if 'M1' in self:
                            curvature = self['M1'].curvature
                        else:
                            curvature = self.m1.curvature
                    
                    position = [
                        offset[0] + shift[0],
                        offset[1] + shift[1],
                        self.m1.sagitta(np.sqrt(offset[0]**2+offset[1]**2)) + shift[2],
                    ]
                    segment = AsphericalSurface(
                        conic_constant=self.m1.conic_constant,
                        aspheric_coefficients=self.m1.aspheric_coefficients,
                        half_aperture=half_aperture, 
                        curvature=curvature,
                        position=position,
                        surface_type=self.m1.type,
                        name = f'M1-segment-{key}',
                        aperture_shape=ApertureShape.HEXAGONAL,
                        tilt_angles=tilt_angles,
                        scattering_dispersion=scattering_dispersion
                    )
                    segment.offset = offset

                    if 'properties' in segment_conf and segment_conf['properties'] is not None:
                        prop_file = segment_conf['properties']
                        surface_properties = SurfaceProperties.load_json(prop_file)
                        segment.properties = surface_properties
                    else:
                        if 'M1' in self:
                            segment.properties = self['M1'].properties

                    self.add_surface(segment, replace=True)
                
                self.remove_surface('M1')

    @staticmethod
    def load_efficiency_data(filepath: str) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Loads surface efficiency data from a specially formatted text file.

        Args
        ----
            filepath: Path to the text file.

        Returns
        -------
            A tuple containing:
            - wavelengths: 1D NumPy array of wavelengths.
            - incidence_angles: 1D NumPy array of incidence angles.
            - efficiencies: 2D NumPy array of efficiencies (rows: angles, cols: wavelengths).

        Notes
        -----
            The input file must be formatted as following::
            
                # Efficiency example
                # Wavelengths (nm)
                300. 350. 450. 500. 550. 600.
                # Incidence angles (degree)
                0. 30. 60. 90.
                # Values
                # A row for each angle
                # Angle 0.
                1. 1. 1. 1. 1. 1.
                # Angle 30.
                0.9 0.9 0.9 0.9 0.9 0.9
                # etc
                0.6 0.6 0.6 0.6 0.6 0.6
                0. 0. 0. 0. 0. 0.

            The first not commented line is interpreted as wavelengths.
            The second not commented line is interpreted as incidence angles.
            The rest of the lines are interpreted as the efficiency values.
            Empty lines are allowed to be present only to indicate that a dependecy is missing::

                # Efficiency example
                # Wavelengths (nm)
                300. 350. 450. 500. 550. 600.
                # Incidence angles
                
                # Values
                1. 1. 1. 1. 1. 1.
            
        
        """
        lines = []
        empties = 0
        with open(filepath, 'r') as f:
            for line in f:
                line = line.strip()
                if not line.startswith('#'):
                    if line == "":
                        empties += 1
                        if empties > 1:
                            raise(RuntimeError(f"{filepath}: at most one empty line is allowed in an efficiency file."))
                    lines.append(line)
        
        wavelengths = np.fromstring(lines[0], sep=' ')
        if wavelengths.shape[0] == 1 and wavelengths[0] == -1:
            wavelengths = None
        
        incidence_angles = np.fromstring(lines[1], sep=' ')
        if incidence_angles.shape[0] == 1 and incidence_angles[0] == -1:
            incidence_angles = None
        
        if incidence_angles is None and wavelengths is None:
            raise(RuntimeError(f"{filepath}: no wavelength or incidence angle found."))

        efficiencies = np.array([np.fromstring(line, sep=' ') for line in lines[2:]])

        return wavelengths, incidence_angles, efficiencies  